/*
 * @(#)ClientAction.java
 *
 * Copyright (c) 2007 masahito suzuki, Inc. All Rights Reserved
 */
package org.maachang.sync ;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.util.Arrays;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.util.ByteUtil;
import org.maachang.util.ConvertBinary;
import org.maachang.util.Encryption;
import org.maachang.util.FileUtil;
import org.maachang.util.net.CreateSocket;
import org.maachang.util.net.ReceiveTimeoutException;
import org.maachang.util.net.Tcp;

/**
 * クライアントアクション処理.
 *
 * @version 2007/05/20
 * @author  masahito suzuki
 * @since   MaachangSync 1.00
 */
public class ClientAction {
    
    /**
     * ログオブジェクト.
     */
    private static final Log LOG = LogFactory.getLog( ClientAction.class ) ;
    
    /**
     * クライアントタイムアウト.
     */
    private static final long CLIENT_TIMEOUT = 60L ;
    
    /**
     * 受信待ちカウント.
     */
    private static final int RECEIVE_ERROR_COUNT = 5000 ;
    
    /**
     * クライアントTCP.
     */
    private Tcp tcp = null ;
    
    /**
     * キーコード.
     */
    private byte[] keycode = null ;
    
    /**
     * ローカルディレクトリ名.
     */
    private String localDir = null ;
    
    /**
     * リモートディレクトリ名.
     */
    private String remoteDir = null ;
    
    /**
     * バイナリバッファ.
     */
    private byte[] buffer = new byte[ MaachangSyncDef.DEFAULT_BINARY ] ;
    
    /**
     * バイナリユーティリティ.
     */
    private ByteUtil byteUtil = new ByteUtil() ;
    
    /**
     * 暗号オブジェクト.
     */
    private Encryption enc = new Encryption() ;
    
    /**
     * コンストラクタ.
     */
    private ClientAction() {
        
    }
    
    /**
     * コンストラクタ.
     * <BR><BR>
     * サーバアクション処理を実施します.
     * <BR>
     * @param socket 対象のソケットを設定します.
     * @param localDir ローカルターゲットディレクトリ名を設定します.
     * @param remoteDir リモートターゲットディレクトリ名を設定します.
     * @param key 対象のキーコードを設定します.
     * @exception Exception 例外.
     */
    public ClientAction( Socket socket,String localDir,String remoteDir,String key )
        throws Exception {
        if( socket == null || localDir == null || localDir.length() <= 0 ||
            remoteDir == null || remoteDir.length() <= 0) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( key == null || key.length() <= 0 ) {
            key = MaachangSyncDef.DEFAULT_KEY ;
        }
        
        if( FileUtil.isFileExists( localDir ) == true ) {
            throw new Exception( "指定ディレクトリ名(" + localDir +
                ")は、ファイル名で存在します" ) ;
        }
        if( FileUtil.isDirExists( localDir ) == false ) {
            FileUtil.mkdirs( localDir ) ;
        }
        this.tcp = CreateSocket.createTcp( socket ) ;
        this.localDir = FileUtil.getFullPath( localDir ) ;
        this.remoteDir = remoteDir ;
        this.keycode = Encryption.convertStringByCode32Key( key ) ;
    }
    
    /**
     * ファイナライズ処理定義.
     * <BR><BR>
     * ファイナライズ処理定義.
     * <BR>
     * @exception Exception 例外処理が返されます.
     */
    protected void finalize() throws Exception {
        try {
            this.clear() ;
        } catch( Exception t ){
        }
    }
    
    /**
     * 処理クリア.
     * <BR><BR>
     * 対象のスレッド処理をクリアします.
     */
    public void clear() {
        if( this.tcp != null ) {
            this.tcp.destroy() ;
        }
        this.tcp = null ;
        this.localDir = null ;
        this.remoteDir = null ;
        this.keycode = null ;
        this.byteUtil = null ;
    }
    
    /**
     * 処理開始.
     * <BR><BR>
     * 同期処理を開始します.
     * <BR>
     * @param opt 対象の実行オプションを行います.<BR>
     *            [true]を設定した場合、全ファイルを総入れ替えします.
     * @exception Exception 例外処理が返されます.
     */
    public void execution( boolean opt ) throws Exception {
        
        // ターゲットディレクトリの最終更新日と、リモートディレクトリ名を送信.
        enc.create( keycode ) ;
        int len = getLastByDirectory( opt,buffer,localDir,remoteDir,enc ) ;
        tcp.send( buffer,0,len ) ;
        tcp.flush() ;
        
        // IPアドレス+ポート番号を取得.
        String ipPort = getIpPort( tcp ) ;
        //LOG.info( "## [" + ipPort + "]接続処理を開始......" ) ;
        
        // ゴミがのこっていないかロールバック.
        rollback( localDir ) ;
        
        // 基本受信処理.
        int[] result = new int[ 1 ] ;
        int cnt = 0 ;
        boolean exitFlg = false ;
        ClientChild clientChild = new ClientChild() ;
        
        try {
            
            // 受信処理.
            for( int i = 0 ;; ) {
                
                // 初期化.
                result[ 0 ] = 0 ;
                
                // １電文取得.
                len = getReceive( result,tcp,buffer,byteUtil ) ;
                if( len <= -1 ) {
                    if( cnt >= RECEIVE_ERROR_COUNT ) {
                        throw new Exception( "受信待ちカウントの制限値を越しました" ) ;
                    }
                    cnt ++ ;
                    continue ;
                }
                cnt = 0 ;
                switch( result[ 0 ] ) {
                    // ファイル受信開始処理.
                    case 1 :
                        // クライアントチャイルドを取得.
                        clientChild.clear() ;
                        enc.create( keycode ) ;
                        getClientChild( ipPort,clientChild,localDir,buffer,len,enc ) ;
                        break ;
                    // ファイル受信処理.
                    case 2 :
                        // ファイル出力.
                        enc.create( keycode ) ;
                        i = getOutputFile( clientChild,buffer,len,enc,i ) ;
                        break ;
                    // 同期終了処理.
                    case 3 :
                        // コミット処理.
                        clientChild.clear() ;
                        commit( localDir ) ;
                        exitFlg = true ;
                        break ;
                }
                
                // 処理終了.
                if( exitFlg == true ) {
                    //LOG.info( "## [" + ipPort + "]受信終了...." ) ;
                    break ;
                }
            }
            
        } catch( Exception e ) {
            //LOG.error( "## [" + ipPort + "]受信中にエラーが発生しました",e ) ;
            // ロールバック処理.
            rollback( localDir ) ;
            throw e ;
        } finally {
            // 処理終了.
            this.clear() ;
            LOG.info( "## [" + ipPort + "]処理終了....." ) ;
        }
    }
    
    /**
     * ターゲットディレクトリ時間を設定
     */
    private static int getLastByDirectory( boolean opt,byte[] buf,String localDir,String remoteDir,Encryption enc )
        throws Exception {
        
        // ディレクトリ存在チェック.
        if( FileUtil.isDirExists( localDir ) == false ) {
            throw new Exception( "指定ディレクトリ名(" + localDir + ")は存在しないか不正です" ) ;
        }
        
        // ターゲットディレクトリ以下の最終更新日を取得.
        long lastUpdate = -1L ;
        if( opt == false ) {
            lastUpdate = FileUtil.getDirectoryToLastTime( localDir ) ;
        }
        
        byte[] fnm = remoteDir.getBytes( MaachangSyncDef.DEFAULT_CHARSET ) ;
        
        // ヘッダ情報.
        buf[ 0 ] = MaachangSyncProtocol.HD_LAST_UPATE[ 0 ] ;
        buf[ 1 ] = MaachangSyncProtocol.HD_LAST_UPATE[ 1 ] ;
        int pnt = 2 ;
        
        // 暗号キー.
        byte[] key = Encryption.getPublicKey() ;
        System.arraycopy( key,0,buf,2,Encryption.ENCRYPION_KEY_LENGTH ) ;
        pnt += Encryption.ENCRYPION_KEY_LENGTH + 1 + 4 ;
        int next = pnt; 
        
        // 最終更新日.
        ConvertBinary.convertLong( buf,pnt,lastUpdate ) ;
        pnt += 8 ;
        
        // パス名長.
        ConvertBinary.convertInt( buf,pnt,fnm.length ) ;
        pnt += 4 ;
        
        // パス名.
        System.arraycopy( fnm,0,buf,pnt,fnm.length ) ;
        pnt += fnm.length ;
        
        // 暗号化.
        int code = enc.encryption( key,buf,next,pnt-next ) ;
        buf[ MaachangSyncProtocol.CHECK_POINT ] = ( byte )( code & 0x000000ff ) ;
        
        // プロトコル長.
        ConvertBinary.convertInt( buf,MaachangSyncProtocol.CHECK_POINT+1,pnt ) ;
        
        return pnt ;
    }
    
    /**
     * 受信処理.
     */
    private static final int getReceive( int[] out,Tcp tcp,byte[] buf,ByteUtil byteUtil )
        throws Exception {
        
        // 受信無限待ち.
        int len = -1 ;
        try {
            len = tcp.receive( buf,-1,CLIENT_TIMEOUT ) ;
            byteUtil.add( buf,0,len ) ;
        } catch( ReceiveTimeoutException rt ) {
            len = -1 ;
        }
        
        int dataLen ;
        if( ( dataLen = byteUtil.size() ) <= MaachangSyncProtocol.DATA_POINT ) {
            return -1 ;
        }
        
        Arrays.fill( buf,( byte )0 ) ;
        byteUtil.get( buf ) ;
        out[ 0 ] = 0 ;
        int ret = -1 ;
        
        // 取得情報を判別.
        if( ( int )( buf[ 0 ] & 0x000000ff ) == ( int )( MaachangSyncProtocol.HD_SYNC_START_FILE[ 0 ] & 0x000000ff ) &&
            ( int )( buf[ 1 ] & 0x000000ff ) == ( int )( MaachangSyncProtocol.HD_SYNC_START_FILE[ 1 ] & 0x000000ff ) ) {
            // ファイル受信開始処理.
            out[ 0 ] = 1 ;
            ret = MaachangSyncProtocol.getProtocolLength( buf ) ;
            if( ret > dataLen ) {
                return -1 ;
            }
        }
        else if(  ( int )( buf[ 0 ] & 0x000000ff ) == ( int )( MaachangSyncProtocol.HD_SYNC_FILE[ 0 ] & 0x000000ff ) &&
            ( int )( buf[ 1 ] & 0x000000ff ) == ( int )( MaachangSyncProtocol.HD_SYNC_FILE[ 1 ] & 0x000000ff ) ) {
            // ファイル受信処理.
            out[ 0 ] = 2 ;
            ret = MaachangSyncProtocol.getProtocolLength( buf ) ;
            if( ret > dataLen ) {
                return -1 ;
            }
        }
        else if(  ( int )( buf[ 0 ] & 0x000000ff ) == ( int )( MaachangSyncProtocol.HD_END[ 0 ] & 0x000000ff ) &&
            ( int )( buf[ 1 ] & 0x000000ff ) == ( int )( MaachangSyncProtocol.HD_END[ 1 ] & 0x000000ff ) ) {
            // 同期終了処理.
            out[ 0 ] = 3 ;
            ret = MaachangSyncProtocol.getProtocolLength( buf ) ;
            if( ret > dataLen ) {
                return -1 ;
            }
        }
        else if( buf[ 0 ] == 0 && buf[ 1 ] == 0 ) {
            return -1 ;
        }
        else {
            throw new Exception( "不正な受信情報です[" + buf[ 0 ] + ":" + buf[ 1 ] + "]" ) ;
        }
        
        // データ長分の内容を破棄.
        byteUtil.remove( ret ) ;
        return ret ;
    }
    
    /**
     * クライアントチャイルドを生成.
     */
    private static final void getClientChild( String ipPort,ClientChild clientChild,String localDir,byte[] buf,int length,Encryption enc )
        throws Exception {
        
        // 暗号解析.
        byte[] key = MaachangSyncProtocol.getKey( buf ) ;
        int code = MaachangSyncProtocol.getCode( buf ) ;
        enc.analysis( key,code,buf,MaachangSyncProtocol.DATA_POINT,
            length-MaachangSyncProtocol.DATA_POINT ) ;
        int pnt = MaachangSyncProtocol.DATA_POINT ;
        
        // 最終更新日.
        int no =  ConvertBinary.convertInt( pnt,buf ) ;
        pnt += 4 ;
        
        // 全ファイルサイズ.
        long all = ConvertBinary.convertLong( pnt,buf ) ;
        pnt += 8 ;
        
        // ターゲットファイル名長.
        int pathLen = ConvertBinary.convertInt( pnt,buf ) ;
        pnt += 4 ;
        if( pathLen <= 0 || pathLen + MaachangSyncProtocol.HD_SYNC_START_FILE_LENGTH > length  ) {
            throw new Exception( "ターゲットファイル名長が不正です[" + pathLen + "]" ) ;
        }
        
        // ターゲットファイル名を取得.
        byte[] nameBin = new byte[ pathLen ] ;
        System.arraycopy( buf,pnt,nameBin,0,pathLen ) ;
        String fileName = new String( nameBin,MaachangSyncDef.DEFAULT_CHARSET ) ;
        String oneFileName = FileUtil.getFileName( fileName ) ;
        nameBin = null ;
        
        // 絶対パスのディレクトリを取得.
        fileName = FileUtil.cutSeparator( true,fileName ) ;
        localDir = FileUtil.cutSeparator( false,localDir ) ;
        String dir = new StringBuffer().append( localDir ).append( FileUtil.FILE_SPACE ).
            append( fileName.substring( 0,fileName.length() - oneFileName.length() ) ).
            toString() ;
        dir = FileUtil.cutSeparator( false,dir ) ;
        dir += FileUtil.FILE_SPACE ;
        
        // ディレクトリを生成.
        if( FileUtil.isDirExists( dir ) == false ) {
            FileUtil.mkdirs( dir ) ;
        }
        
        // テンポラリファイル名で生成.
        BufferedOutputStream bo = new BufferedOutputStream(
            new FileOutputStream( dir+MaachangSyncDef.TEMP_FILENAME+oneFileName ) ) ;
        
        // 新しいチャイルド条件を生成.
        clientChild.setTargetName( dir+oneFileName ) ;
        clientChild.setTargetFileLength( all ) ;
        clientChild.setFileNo( no ) ;
        clientChild.setTargetFp( bo ) ;
        
        // KByte条件に変換.
        long kb = clientChild.getTargetFileLength() / 1024L ;
        if( kb <= 0L ) {
            kb = 1L ;
        }
        
        // 受信状況表示.
        System.out.println() ;
        LOG.info( "## [" + ipPort + "]" + clientChild.getTargetName() + "(" + kb + "KByte)" ) ;
        
    }
    
    /**
     * ファイル出力.
     */
    private static final int getOutputFile( ClientChild clientChild,byte[] buf,int length,Encryption enc,int seq )
        throws Exception {
        
        // 暗号解析.
        byte[] key = MaachangSyncProtocol.getKey( buf ) ;
        int code = MaachangSyncProtocol.getCode( buf ) ;
        enc.analysis( key,code,buf,MaachangSyncProtocol.DATA_POINT,
            length-MaachangSyncProtocol.DATA_POINT ) ;
        int pnt = MaachangSyncProtocol.DATA_POINT ;
        
        // ファイル項番.
        int no =  ConvertBinary.convertInt( pnt,buf ) ;
        pnt += 4 ;
        
        // ファイル項番が一致しない.
        if( no != clientChild.getFileNo() ) {
            throw new Exception( "ファイル名[" + clientChild.getTargetName() +
                "]の項番[" + clientChild.getFileNo() + "/" + no + "]が一致しません" ) ;
        }
        
        // ファイル項番.
        pnt += 4 ;
        
        // 送信サイズ.
        int sendLen = ConvertBinary.convertInt( pnt,buf ) ;
        pnt += 4 ;
        
        // 送信サイズと一致しない条件の場合.
        if( sendLen <= 0 || length - MaachangSyncProtocol.HD_SYNC_FILE_LENGTH != sendLen ) {
            throw new Exception( "ファイル名[" + clientChild.getTargetName() +
                "]の受信サイズ[" + sendLen + "]が不正です" ) ;
        }
        
        // ファイル書き込み.
        OutputStream os = clientChild.getTargetFp() ;
        os.write( buf,MaachangSyncProtocol.HD_SYNC_FILE_LENGTH,sendLen ) ;
        
        // 受信状況表示.
        if( seq >= 5 ) {
            System.out.print( "." ) ;
            seq = 0 ;
        }
        else {
            seq ++ ;
        }
        return seq ;
    }
    
    /**
     * コミット処理.
     */
    private static final void commit( String localDir )
        throws Exception {
        commitToDir( localDir ) ;
    }
    
    /**
     * ロールバック処理.
     */
    private static final void rollback( String localDir )
        throws Exception {
        rollbackToDir( localDir ) ;
    }
    
    /**
     * コミットをディレクトリ単位で処理.
     */
    private static final void commitToDir( String name )
        throws Exception {
        
        if( name.endsWith( "\\" ) == false && name.endsWith( "/" ) == false ) {
            name = name + FileUtil.FILE_SPACE ;
        }
        
        File file = new File( name ) ;
        String[] names = file.list() ;
        file = null ;
        int len ;
        if( names != null && ( len = names.length ) > 0 ) {
            for( int i = 0 ; i < len ; i ++ ) {
                String target = name + names[ i ] ;
                names[ i ] = null ;
                File fp = new File( target ) ;
                
                if( fp.isDirectory() == true ) {
                    fp = null ;
                    commitToDir( target ) ;
                }
                else {
                    fp = null ;
                    if( FileUtil.getFileName( target ).startsWith( MaachangSyncDef.TEMP_FILENAME ) == true ) {
                        String fileName = FileUtil.getFileName( target ) ;
                        String destName = new StringBuffer().
                            append( target.substring( 0,target.length() - ( fileName.length() + 1 ) ) ).
                            append( FileUtil.FILE_SPACE ).
                            append( fileName.substring( MaachangSyncDef.TEMP_FILENAME.length() ) ).toString() ;
                        
                        if( FileUtil.isFileExists( destName ) == true ) {
                            FileUtil.removeFile( destName ) ;
                        }
                        FileUtil.moveFile( target,destName ) ;
                    }
                }
            }
        }
        
    }
    
    /**
     * ロールバックをディレクトリ単位で処理.
     */
    private static final void rollbackToDir( String name )
        throws Exception {
        
        if( name.endsWith( "\\" ) == false && name.endsWith( "/" ) == false ) {
            name = name + FileUtil.FILE_SPACE ;
        }
        
        File file = new File( name ) ;
        String[] names = file.list() ;
        file = null ;
        int len ;
        if( names != null && ( len = names.length ) > 0 ) {
            for( int i = 0 ; i < len ; i ++ ) {
                String target = name + names[ i ] ;
                names[ i ] = null ;
                File fp = new File( target ) ;
                
                if( fp.isDirectory() == true ) {
                    fp = null ;
                    rollbackToDir( target ) ;
                }
                else {
                    fp = null ;
                    if( FileUtil.getFileName( target ).startsWith( MaachangSyncDef.TEMP_FILENAME ) == true ) {
                        if( FileUtil.isFileExists( target ) == true ) {
                            FileUtil.removeFile( target ) ;
                        }
                    }
                }
            }
        }
        
    }
    
    /**
     * IPアドレスとポート番号を文字で取得.
     */
    private static final String getIpPort( Tcp tcp ) {
        String ret = null ;
        if( tcp == null || tcp.isTcp() == false ) {
            ret = "unknown/xxxxxx" ;
        }
        else {
            Socket sock = tcp.getSocket() ;
            ret = new StringBuffer().
                append( sock.getInetAddress().getHostAddress() ).
                append( "/" ).
                append( sock.getPort() ).toString() ;
        }
        return ret ;
    }
}
