/*
 * @(#)ClientMain.java
 *
 * Copyright (c) 2007 masahito suzuki, Inc. All Rights Reserved
 */
package org.maachang.sync ;

import java.net.InetAddress;
import java.net.Socket;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.util.net.CreateSocket;

/**
 * サーバメイン処理.
 *
 * @version 2007/05/20
 * @author  masahito suzuki
 * @since   MaachangSync 1.00
 */
public class ClientMain {
    
    /**
     * ログオブジェクト.
     */
    private static final Log LOG = LogFactory.getLog( ClientMain.class ) ;
    
    /**
     * パラメータ : ローカルディレクトリ.
     */
    private static final String PARAM_LOCAL_DIR = "-o" ;
    
    /**
     * パラメータ : リモートディレクトリ.
     */
    private static final String PARAM_REMOTE_DIR = "-d" ;
    
    /**
     * パラメータ : 暗号キーワード.
     */
    private static final String PARAM_KEY = "-k" ;
    
    /**
     * パラメータ : リモートアドレス.
     */
    private static final String PARAM_REMOTE = "-r" ;
    
    /**
     * パラメータ : ポート番号.
     */
    private static final String PARAM_PORT = "-p" ;
    
    /**
     * パラメータ : 総置き換えオプション.
     */
    private static final String PARAM_ALL = "-all" ;
    
    /**
     * メイン処理.
     */
    public static final void main( String[] args )
        throws Exception {
        new ClientMain().execution( args ) ;
    }
    
    /**
     * クライアントメイン実行.
     * <BR><BR>
     * クライアントメイン実行処理.
     * <BR>
     * @param args プロセス引数を設定します.
     * @exception Exception 例外.
     */
    public void execution( String[] args )
        throws Exception {
        
        // パラメータを取得.
        String localDir = getLocalDir( args ) ;
        String remoteDir = getRemoteDir( args ) ;
        String keyword = getKeyword( args ) ;
        String remoteAddr = getRemote( args ) ;
        int port = getPort( args ) ;
        boolean allOpt = isAllOption( args ) ;
        
        if( localDir == null || localDir.length() <= 0 ||
            remoteDir == null || remoteDir.length() <= 0 ||
            remoteAddr == null || remoteAddr.length() <= 0 ) {
            LOG.error( "## 引数が足りません" ) ;
            return ;
        }
        
        // 開始ログ.
        LOG.info( "## クライアント開始 MaachangSync#ClientMain[" +
            " localDir:" + localDir + " remoteDir:" + remoteDir +
            " keyword:" + keyword + " remoteAddr:" + remoteAddr +
            " port:" + port + " allOpt:" + allOpt + "]" ) ;
        
        try {
            
            // ソケットオープン.
            Socket socket = CreateSocket.clientSocket( InetAddress.getByName( remoteAddr ),port ) ;
            
            // オープンログ.
            LOG.info( "## addr:" + socket.getLocalAddress().getHostAddress() +
                " port:" + socket.getLocalPort() ) ;
            
            // クライアントアクション.
            ClientAction cl = new ClientAction( socket,localDir,remoteDir,keyword ) ;
            cl.execution( allOpt ) ;
            
        } catch( Exception e ) {
            LOG.error( "## クライアントエラー",e ) ;
        } finally {
            LOG.info( "## クライアント終了....." ) ;
        }
        
    }
    
    /**
     * ローカルディレクトリを取得.
     */
    private static final String getLocalDir( String[] args ) {
        if( args == null ) {
            return null ;
        }
        int len = args.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            String name = args[ i ] ;
            if( name == null || ( name = name.trim() ).length() <= 0 ) {
                continue ;
            }
            if( name.toLowerCase().startsWith( PARAM_LOCAL_DIR ) == true ) {
                return name.substring( PARAM_LOCAL_DIR.length() ) ;
            }
        }
        return null ;
    }
    
    /**
     * リモートディレクトリを取得.
     */
    private static final String getRemoteDir( String[] args ) {
        if( args == null ) {
            return null ;
        }
        int len = args.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            String name = args[ i ] ;
            if( name == null || ( name = name.trim() ).length() <= 0 ) {
                continue ;
            }
            if( name.toLowerCase().startsWith( PARAM_REMOTE_DIR ) == true ) {
                return name.substring( PARAM_REMOTE_DIR.length() ) ;
            }
        }
        return null ;
    }
    
    /**
     * キーワードを取得.
     */
    private static final String getKeyword( String[] args ) {
        if( args == null ) {
            return MaachangSyncDef.DEFAULT_KEY ;
        }
        int len = args.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            String name = args[ i ] ;
            if( name == null || ( name = name.trim() ).length() <= 0 ) {
                continue ;
            }
            if( name.toLowerCase().startsWith( PARAM_KEY ) == true ) {
                return name.substring( PARAM_KEY.length() ) ;
            }
        }
        return MaachangSyncDef.DEFAULT_KEY ;
    }
    
    /**
     * リモートアドレスを取得.
     */
    private static final String getRemote( String[] args ) {
        if( args == null ) {
            return null ;
        }
        int len = args.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            String name = args[ i ] ;
            if( name == null || ( name = name.trim() ).length() <= 0 ) {
                continue ;
            }
            if( name.toLowerCase().startsWith( PARAM_REMOTE ) == true ) {
                return name.substring( PARAM_REMOTE.length() ) ;
            }
        }
        return null ;
    }
    
    /**
     * ポート番号を取得.
     */
    private static final int getPort( String[] args ) {
        if( args == null ) {
            return MaachangSyncDef.DEFAULT_PORT ;
        }
        int len = args.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            String name = args[ i ] ;
            if( name == null || ( name = name.trim() ).length() <= 0 ) {
                continue ;
            }
            if( name.toLowerCase().startsWith( PARAM_PORT ) == true ) {
                String port = name.substring( PARAM_PORT.length() ) ;
                try {
                    int portInt = Integer.parseInt( port ) ;
                    if( portInt >= 0 && portInt < 65535 ) {
                        return portInt ;
                    }
                } catch( Exception e ) {
                }
            }
        }
        return MaachangSyncDef.DEFAULT_PORT ;
    }
    
    /**
     * 総置き換えオプションを取得.
     */
    private static final boolean isAllOption( String[] args ) {
        if( args == null ) {
            return false ;
        }
        int len = args.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            String name = args[ i ] ;
            if( name == null || ( name = name.trim() ).length() <= 0 ) {
                continue ;
            }
            if( name.toLowerCase().equals( PARAM_ALL ) == true ) {
                return true ;
            }
        }
        return false ;
    }
}
