/*
 * @(#)ServerAction.java
 *
 * Copyright (c) 2007 masahito suzuki, Inc. All Rights Reserved
 */
package org.maachang.sync ;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.net.Socket;
import java.util.ArrayList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.util.ConvertBinary;
import org.maachang.util.Encryption;
import org.maachang.util.FileUtil;
import org.maachang.util.net.CreateSocket;
import org.maachang.util.net.Tcp;

/**
 * サーバアクション処理.
 *
 * @version 2007/05/19
 * @author  masahito suzuki
 * @since   MaachangSync 1.00
 */
public class ServerAction extends Thread {
    
    /**
     * ログオブジェクト.
     */
    private static final Log LOG = LogFactory.getLog( ServerAction.class ) ;
    
    /**
     * １回の送信バッファ.
     */
    private static final int ONE_SEND_LENGTH = MaachangSyncDef.DEFAULT_BINARY-MaachangSyncProtocol.HD_SYNC_FILE_LENGTH ;
    
    /**
     * TCP処理.
     */
    private Tcp tcp = null ;
    
    /**
     * キーコード.
     */
    private byte[] keycode = null ;
    
    /**
     * 終了フラグ.
     */
    private volatile boolean endFlg = false ;
    
    /**
     * バイナリバッファ.
     */
    private byte[] buffer = new byte[ MaachangSyncDef.DEFAULT_BINARY ] ;
    
    /**
     * 暗号オブジェクト.
     */
    private Encryption enc = new Encryption() ;
    
    /**
     * 同期オブジェクト.
     */
    private final Object sync = new Object() ;
    
    /**
     * コンストラクタ.
     */
    private ServerAction() {
        
    }
    
    /**
     * コンストラクタ.
     * <BR><BR>
     * サーバアクション処理を実施します.
     * <BR>
     * @param socket 対象のソケットを設定します.
     * @param key 対象のキーコードを設定します.
     * @exception Exception 例外.
     */
    public ServerAction( Socket socket,String key )
        throws Exception {
        if( socket == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( key == null || key.length() <= 0 ) {
            key = MaachangSyncDef.DEFAULT_KEY ;
        }
        this.tcp = CreateSocket.createTcp( socket ) ;
        this.keycode = Encryption.convertStringByCode32Key( key ) ;
        this.checkAccess() ;
        this.setDaemon( false ) ;
        endFlg = false ;
        this.start() ;
    }
    
    /**
     * ファイナライズ処理定義.
     * <BR><BR>
     * ファイナライズ処理定義.
     * <BR>
     * @exception Exception 例外処理が返されます.
     */
    protected void finalize() throws Exception {
        try {
            this.clear() ;
        } catch( Exception t ){
        }
    }
    
    /**
     * 処理クリア.
     * <BR><BR>
     * 対象のスレッド処理をクリアします.
     */
    public void clear() {
        synchronized( sync ) {
            this.endFlg = true ;
            if( this.tcp != null ) {
                this.tcp.destroy() ;
            }
            this.tcp = null ;
        }
    }
    
    /**
     * スレッド動作チェック.
     * <BR><BR>
     * スレッド動作チェックします.
     * <BR>
     * @return boolean スレッド動作チェックが返されます.
     */
    public boolean isThread() {
        boolean ret = true ;
        synchronized( sync ) {
            ret = ( endFlg == true ) ? false : true ;
        }
        return ret ;
    }
    
    /**
     * スレッド処理実施.
     * <BR><BR>
     * スレッド処理実施を行います.
     */
    public void run() {
        String ipPort = this.getIpPort() ;
        synchronized( sync ) {
            this.endFlg = false ;
            if( tcp.isTcp() == false ) {
                LOG.error( "## コネクションに失敗....." ) ;
                return ;
            }
        }
        
        //LOG.info( "## [" + ipPort + "]同期処理を開始....." ) ;
        
        try {
            this.execution( ipPort ) ;
        } catch( Exception e ) {
            LOG.error( "## [" + ipPort + "]同期処理でエラー.....",e ) ;
        } finally {
            //LOG.info( "## [" + ipPort + "]同期処理を終了....." ) ;
            synchronized( sync ) {
                this.endFlg = true ;
                if( this.tcp != null ) {
                    this.tcp.destroy() ;
                }
                this.tcp = null ;
            }
        }
    }
    
    /**
     * IPアドレスとポート番号を文字で取得.
     */
    private final String getIpPort() {
        String ret = null ;
        synchronized( sync ) {
            ret = getIpPort( tcp ) ;
        }
        return ret ;
    }
    
    /**
     * サーバスレッド実行処理.
     */
    private final void execution( String ipPort )
        throws Exception {
        ArrayList lst = null ;
        
        Tcp tcpip = null ;
        byte[] key = null ;
        synchronized( sync ) {
            tcpip = this.tcp ;
            key = this.keycode ;
        }
        
        // クライアントの最終更新日を取得.
        String[] targetDir = new String[ 1 ] ;
        long time = ServerAction.getLastUpdateTime( targetDir,tcpip,key,buffer,enc ) ;
        String target = FileUtil.getFullPath( targetDir[ 0 ] ) ;
        targetDir = null ;
        
        LOG.info( "## [" + ipPort + "](" + target + ")=クライアントの最終更新時間-" + time + "sec..." ) ;
        
        if( FileUtil.isDirExists( target ) == false ) {
            LOG.error( "リモート指定のディレクトリ名(" + target + ")は存在しないか不正です" ) ;
            throw new Exception( "リモート指定のディレクトリ名(" + target + ")は存在しないか不正です" ) ;
        }
        
        // 送信対象ファイル群を取得.
        lst = ServerAction.getSendFileNames( target,time ) ;
        
        if( lst == null || lst.size() <= 0 ) {
            LOG.info( "## [" + ipPort + "](" + target + ")=送信ファイル総数-なし" ) ;
            ServerAction.exitSend( tcpip,key,0,buffer,enc ) ;
            return ;
        }
        
        LOG.info( "## [" + ipPort + "](" + target + ")=送信ファイル総数-" + lst.size() + "個" ) ;
        
        int len = lst.size() ;
        for( int i = 0 ; i < len ; i ++ ) {
            String syncFile = ( String )lst.get( i ) ;
            //LOG.info( "## -- [" + ipPort + "](" + target + "," + i + ")=送信ファイル名の同期開始-" + syncFile ) ;
            LOG.info( "## -- [" + ipPort + "](" + target + "," + i + ")=送信ファイル名の同期-" + syncFile ) ;
            // １つのファイルを送信.
            ServerAction.sendFile( tcpip,key,target,syncFile,i,buffer,enc ) ;
            // ファイルクローズの場合は、中断.
            synchronized( sync ) {
                if( endFlg == true ) {
                    LOG.info( "## ++ [" + ipPort + "]("  + target + "," + i + ")=送信ファイル名の同期異常終了-" + syncFile ) ;
                    break ;
                }
            }
            //LOG.info( "## ** [" + ipPort + "]("  + target + "," + i + ")=送信ファイル名の同期正常終了-" + syncFile ) ;
        }
        
        // ファイル送信中に中断した場合.
        boolean rollback = false ;
        synchronized( sync ) {
            if( endFlg == true ) {
                rollback = true ;
            }
        }
        
        // ファイル送信が正常に行われた場合.
        if( rollback == false ) {
            ServerAction.exitSend( tcpip,key,len,buffer,enc ) ;
            LOG.info( "## [" + ipPort + "](" + target + ")=送信ファイル正常同期完了-" + lst.size() ) ;
        }
        else {
            LOG.error( "## [" + ipPort + "](" + target + ")=送信ファイル異常同期完了-" + lst.size() ) ;
        }
    }
    
    /**
     * IPアドレスとポート番号を文字で取得.
     */
    private static final String getIpPort( Tcp tcp ) {
        String ret = null ;
        if( tcp == null || tcp.isTcp() == false ) {
            ret = "unknown/xxxxxx" ;
        }
        else {
            Socket sock = tcp.getSocket() ;
            ret = new StringBuffer().
                append( sock.getInetAddress().getHostAddress() ).
                append( "/" ).
                append( sock.getPort() ).toString() ;
        }
        return ret ;
    }
    
    /**
     * クライアント最終更新日を取得.
     */
    private static final long getLastUpdateTime( String[] out,Tcp tcp,byte[] keycode,byte[] buf,Encryption enc )
        throws Exception {
        
        // 受信処理.
        int len = tcp.receive( buf,-1,MaachangSyncDef.ACCEPT_TIMEOUT ) ;
        if( len <= MaachangSyncProtocol.HD_LAST_UPDATE_LENGTH ) {
            throw new Exception( "最終更新日の受信長が不正です[" + len + "]" ) ;
        }
        
        // ヘッダ確認.
        byte[] header = MaachangSyncProtocol.getHeader( buf ) ;
        if( ( int )( header[ 0 ] & 0x000000ff ) != ( int )( MaachangSyncProtocol.HD_LAST_UPATE[ 0 ] & 0x000000ff ) &&
            ( int )( header[ 1 ] & 0x000000ff ) != ( int )( MaachangSyncProtocol.HD_LAST_UPATE[ 1 ] & 0x000000ff ) ) {
            throw new Exception( "最終更新日のヘッダではありません" ) ;
        }
        
        // 暗号解析.
        byte[] key = MaachangSyncProtocol.getKey( buf ) ;
        int code = MaachangSyncProtocol.getCode( buf ) ;
        enc.create( keycode ) ;
        enc.analysis( key,code,buf,MaachangSyncProtocol.DATA_POINT,
            len-MaachangSyncProtocol.DATA_POINT ) ;
        
        // 最終更新日.
        int pnt = MaachangSyncProtocol.DATA_POINT ;
        long ret =  ConvertBinary.convertLong( pnt,buf ) ;
        pnt += 8 ;
        
        // ターゲットディレクトリ名長.
        int targetLen = ConvertBinary.convertInt( pnt,buf ) ;
        pnt += 4 ;
        if( targetLen <= 0 || targetLen + MaachangSyncProtocol.HD_LAST_UPDATE_LENGTH > len  ) {
            throw new Exception( "ターゲットディレクトリ名長が不正です[" + targetLen + "]" ) ;
        }
        
        // ターゲットディレクトリ名を取得.
        byte[] nameBin = new byte[ targetLen ] ;
        System.arraycopy( buf,pnt,nameBin,0,targetLen ) ;
        out[ 0 ] = new String( nameBin,MaachangSyncDef.DEFAULT_CHARSET ) ;
        nameBin = null ;
        
        // 最終更新日を返す.
        return ret ;
    }
    
    /**
     * 送信対象ファイル群を取得.
     */
    private static final ArrayList getSendFileNames( String targetDir,long lastTime )
        throws Exception {
        ArrayList ret = new ArrayList() ;
        FileUtil.getDirectoryToTargetNextTime( ret,targetDir,lastTime ) ;
        if( ret.size() <= 0 ) {
            return null ;
        }
        return ret ;
    }
    
    /**
     * １つのファイルを送信.
     */
    private static final void sendFile( Tcp tcp,byte[] keycode,String targetDir,String sendName,int no,byte[] buf,Encryption enc )
        throws Exception {
        
        BufferedInputStream bi = null ;
        
        try {
            
            // ファイルヘッダプロトコルを送信.
            long fsz = FileUtil.getLength( sendName ) ;
            enc.create( keycode ) ;
            int len = getFirstSendFile( buf,targetDir,sendName,fsz,no,enc ) ;
            tcp.send( buf,0,len ) ;
            tcp.flush() ;
            
            // ファイルをオープン.
            bi = new BufferedInputStream( new FileInputStream( sendName ) ) ;
            
            // ファイル送信.
            for( int i = 0 ;; i ++ ) {
                enc.create( keycode ) ;
                int one = getSyncOneFile( buf,bi,no,i,fsz,enc ) ;
                fsz -= ( long )( one - MaachangSyncProtocol.HD_SYNC_FILE_LENGTH ) ;
                tcp.send( buf,0,one ) ;
                tcp.flush() ;
                if( fsz <= 0L ) {
                    break ;
                }
            }
            
        } catch( Exception e ) {
            throw e ;
        } finally {
            // ファイルクローズ.
            if( bi != null ) {
                try {
                    bi.close() ;
                } catch( Exception ee ) {
                }
            }
            bi = null ;
        }
        
    }
    
    /**
     * 送信終了を通知.
     */
    private static final void exitSend( Tcp tcp,byte[] keycode,int length,byte[] buf,Encryption enc )
        throws Exception {
        
        enc.create( keycode ) ;
        int len = getEndSendFile( buf,length,enc ) ;
        tcp.send( buf,0,len ) ;
        tcp.flush() ;
        
    }
    
    
    
    /**
     * ファイルヘッダ送信情報を生成.
     */
    private static final int getFirstSendFile( byte[] buf,String targetDir,String sendName,long fsz,int no,Encryption enc )
        throws Exception {
        int pnt = 0 ;
        
        String name = sendName.substring( targetDir.length() ) ;
        if( name.startsWith( "/" ) == true || name.startsWith( "\\" ) == true ) {
            name = name.substring( 1 ) ;
        }
        byte[] fnm = name.getBytes( MaachangSyncDef.DEFAULT_CHARSET ) ;
        
        // ヘッダ情報.
        buf[ 0 ] = MaachangSyncProtocol.HD_SYNC_START_FILE[ 0 ] ;
        buf[ 1 ] = MaachangSyncProtocol.HD_SYNC_START_FILE[ 1 ] ;
        pnt += 2 ;
        
        // 暗号キー.
        byte[] key = Encryption.getPublicKey() ;
        System.arraycopy( key,0,buf,2,Encryption.ENCRYPION_KEY_LENGTH ) ;
        pnt += Encryption.ENCRYPION_KEY_LENGTH + 1 + 4 ;
        int next = pnt; 
        
        // ファイル項番.
        ConvertBinary.convertInt( buf,pnt,no ) ;
        pnt += 4 ;
        
        // 全ファイルサイズ.
        ConvertBinary.convertLong( buf,pnt,fsz ) ;
        pnt += 8 ;
        
        // パス名長.
        ConvertBinary.convertInt( buf,pnt,fnm.length ) ;
        pnt += 4 ;
        
        // パス名.
        System.arraycopy( fnm,0,buf,pnt,fnm.length ) ;
        pnt += fnm.length ;
        
        // 暗号化.
        int code = enc.encryption( key,buf,next,pnt-next ) ;
        buf[ MaachangSyncProtocol.CHECK_POINT ] = ( byte )( code & 0x000000ff ) ;
        
        // プロトコル長.
        ConvertBinary.convertInt( buf,MaachangSyncProtocol.CHECK_POINT+1,pnt ) ;
        
        return pnt ;
    }
    
    /**
     * １回で送信するファイル内容を生成.
     */
    private static final int getSyncOneFile( byte[] buf,InputStream bi,int no,int seq,long all,Encryption enc )
        throws Exception {
        int pnt = 0 ;
        
        // ヘッダ情報.
        buf[ 0 ] = MaachangSyncProtocol.HD_SYNC_FILE[ 0 ] ;
        buf[ 1 ] = MaachangSyncProtocol.HD_SYNC_FILE[ 1 ] ;
        pnt += 2 ;
        
        // 暗号キー.
        byte[] key = Encryption.getPublicKey() ;
        System.arraycopy( key,0,buf,2,Encryption.ENCRYPION_KEY_LENGTH ) ;
        pnt += Encryption.ENCRYPION_KEY_LENGTH + 1 + 4 ;
        int next = pnt; 
        
        // ファイル項番.
        ConvertBinary.convertInt( buf,pnt,no ) ;
        pnt += 4 ;
        
        // シーケンスNo.
        ConvertBinary.convertInt( buf,pnt,seq ) ;
        pnt += 4 ;
        
        // 今回送信サイズ.
        int oneLen = ( all > ( long )ONE_SEND_LENGTH ) ?
            ONE_SEND_LENGTH : ( int )all ;
        ConvertBinary.convertInt( buf,pnt,oneLen ) ;
        pnt += 4 ;
        
        // 送信情報.
        bi.read( buf,pnt,oneLen ) ;
        pnt += oneLen ;
        
        // 暗号化.
        int code = enc.encryption( key,buf,next,pnt-next ) ;
        buf[ MaachangSyncProtocol.CHECK_POINT ] = ( byte )( code & 0x000000ff ) ;
        
        // プロトコル長.
        ConvertBinary.convertInt( buf,MaachangSyncProtocol.CHECK_POINT+1,pnt ) ;
        
        return pnt ;
    }
    
    /**
     * ファイルヘッダ送信情報を生成.
     */
    private static final int getEndSendFile( byte[] buf,int length,Encryption enc )
        throws Exception {
        int pnt = 0 ;
        
        // ヘッダ情報.
        buf[ 0 ] = MaachangSyncProtocol.HD_END[ 0 ] ;
        buf[ 1 ] = MaachangSyncProtocol.HD_END[ 1 ] ;
        pnt += 2 ;
        
        // 暗号キー.
        byte[] key = Encryption.getPublicKey() ;
        System.arraycopy( key,0,buf,2,Encryption.ENCRYPION_KEY_LENGTH ) ;
        pnt += Encryption.ENCRYPION_KEY_LENGTH + 1 + 4 ;
        int next = pnt; 
        
        // 送信ファイル数.
        ConvertBinary.convertInt( buf,pnt,length ) ;
        pnt += 4 ;
        
        // 暗号化.
        int code = enc.encryption( key,buf,next,pnt-next ) ;
        buf[ MaachangSyncProtocol.CHECK_POINT ] = ( byte )( code & 0x000000ff ) ;
        
        // プロトコル長.
        ConvertBinary.convertInt( buf,MaachangSyncProtocol.CHECK_POINT+1,pnt ) ;
        
        return pnt ;
    }

}
