/*
 * @(#)Tcp.java
 *
 * Copyright (c) 2007 masahito suzuki, Inc. All Rights Reserved
 */
package org.maachang.util.net ;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.net.Socket;

/**
 * TCP/IPオブジェクト.
 *  
 * @version 2007/05/19
 * @author  masahito suzuki
 * @since   MaachangSync 1.00
 */
public class Tcp {
    
    /**
     * ソケットオブジェクト.
     */
    private Socket socket = null ;
    
    /**
     * 送信バッファ.
     */
    private BufferedOutputStream sendBuffer = null ;
    
    /**
     * 受信バッファ.
     */
    private BufferedInputStream receiveBuffer = null ;
    
    /**
     * コンストラクタ.
     */
    private Tcp() {
        
    }
    
    /**
     * コンストラクタ.
     * <BR><BR>
     * ソケットオブジェクトを設定して、オブジェクトを
     * 生成します.
     * <BR>
     * @param socket 対象のソケットオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public Tcp( Socket socket )
        throws Exception {
        
        if( socket == null || socket.isClosed() == true ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        
        this.socket = socket ;
        this.sendBuffer = new BufferedOutputStream(
            socket.getOutputStream() ) ;
        this.receiveBuffer = new BufferedInputStream(
            socket.getInputStream() ) ;
    }
    
    /**
     * オブジェクト破棄.
     * <BR><BR>
     * オブジェクトを破棄します.
     * <BR>
     * @exception Exception 例外.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクトを破棄.
     * <BR><BR>
     * オブジェクトを破棄します.
     */
    public void destroy() {
        if( socket != null ) {
            try {
                sendBuffer.close() ;
            } catch( Exception e ) {
            }
            try {
                receiveBuffer.close() ;
            } catch( Exception e ) {
            }
            try {
                socket.close() ;
            } catch( Exception e ) {
            }
        }
        socket = null ;
        sendBuffer = null ;
        receiveBuffer = null ;
    }
    
    /**
     * ソケットオブジェクトを取得.
     * <BR><BR>
     * ソケットオブジェクトを取得します.
     * <BR>
     * @return Socket ソケットオブジェクトが返されます.
     */
    public Socket getSocket() {
        return socket ;
    }
    
    /**
     * InputStreamを取得.
     * <BR><BR>
     * InputStreamを取得します.
     * <BR>
     * @return InputStream 返されます.
     */
    public InputStream getInputStream() {
        return receiveBuffer ;
    }
    
    /**
     * OutputStreamを取得.
     * <BR><BR>
     * OutputStreamを取得します.
     * <BR>
     * @return OutputStream 返されます.
     */
    public OutputStream getOutputStream() {
        return sendBuffer ;
    }
    
    /**
     * 送信情報を更新.
     * <BR><BR>
     * 送信情報を更新します.
     * <BR>
     * @exception Exception 例外.
     */
    public void flush() throws Exception {
        if( sendBuffer != null ) {
            sendBuffer.flush() ;
        }
    }
    
    /**
     * 送信処理.
     * <BR><BR>
     * 対象バイナリの情報を送信します.
     * <BR>
     * @param binary 送信対象のバイナリを設定します.
     * @param offset 対象のオフセット値を設定します.
     * @param length 対象の送信長を設定します.
     * @exception Exception 例外.
     */
    public void send( byte[] binary,int offset,int length )
        throws Exception {
        
        int len ;
        if( binary == null || ( len = binary.length ) <= 0 || offset >= len ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        
        if( offset <= 0 ) {
            offset = 0 ;
        }
        if( length <= 0 || length >= len ) {
            length = len ;
        }
        if( offset + length >= len ) {
            length = len - offset ;
        }
        
        sendBuffer.write( binary,offset,length ) ;
    }
    
    /**
     * 受信処理.
     * <BR><BR>
     * 受信処理を行います.
     * <BR>
     * @param binary 受信情報を格納するバイナリを設定します.
     * @param length 受信情報長を設定します.
     * @param timeout 受信タイムアウト値を設定します.
     * @return int 受信データ長が返されます.
     * @exception ReceiveTimeoutException 受信タイムアウト例外.
     * @exception Exception 例外.
     */
    public int receive( byte[] binary,int length,long timeout )
        throws ReceiveTimeoutException,Exception {
        
        int len ;
        if( binary == null || ( len = binary.length ) <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        
        if( length <= 0 || length >= len ) {
            length = len ;
        }
        
        int ret = 0 ;
        long tm = -1L ;
        if( timeout > 0 ) {
            tm = System.currentTimeMillis() + timeout ;
        }
        
        boolean success = false ;
        for( ;; ) {
            if( tm > -1L && tm <= System.currentTimeMillis() ) {
                throw new ReceiveTimeoutException( "タイムアウトを検知" ) ;
            }
            try {
                ret = receiveBuffer.read( binary,0,length ) ;
                success = true ;
            } catch( InterruptedIOException ii ) {
                ret = -1 ;
                success = false ;
            } catch( Exception e ) {
                throw e ;
            }
            if( success == true ) {
                break ;
            }
        }
        
        return ret ;
    }
    
    /**
     * TCP/IPが有効かチェック.
     * <BR><BR>
     * TCP/IPが有効であるかチェックします.
     * <BR>
     * @return boolean チェック結果が返されます.
     */
    public boolean isTcp() {
        boolean ret = false ;
        try {
            ret = socket.isClosed() ;
            if( ret == true ) {
                ret = false ;
            }
            else {
                ret = true ;
            }
        } catch( Exception e ) {
            ret = false ;
        }
        
        return ret ;
    }
}

